/*!
 * \file
 * \brief Collective benchmark
 * \details Checks and benchmarks collective operations
 * 
 * \b collective.plot draws the graph of the execution time (sec) against message size (kb) with error bars.
 * - input: collective.out
 * - output: collective.eps
 */

#include "mpib.h"
#include <string.h>
#include <getopt.h>
#include <stdio.h>

int main(int argc, char** argv)
{
	MPI_Init(&argc, &argv);
	MPI_Comm comm = MPI_COMM_WORLD;
#ifndef MPIB_DEBUG
	MPIB_Comm(MPI_COMM_WORLD, &comm);
	if (comm == MPI_COMM_NULL)
		goto exit;
#endif

	int exit;
	MPIB_getopt_help_default(&exit);
	char operation[256];
	char timing[256];
	MPIB_getopt_coll_default(operation, timing);
	int stride;
	int max_size;
	MPIB_getopt_msg_default(&stride, &max_size);
	MPIB_precision precision;
	MPIB_getopt_precision_default(&precision);

	int rank;
	MPI_Comm_rank(comm, &rank);
	if (rank == 0)
	{
		int c;
		char options[256] = "";
		strcat(options, MPIB_getopt_help_options());
		strcat(options, MPIB_getopt_coll_options());
		strcat(options, MPIB_getopt_msg_options());
		strcat(options, MPIB_getopt_precision_options());
		while ((c = getopt(argc, argv, options)) != -1)
		{
			MPIB_getopt_help_optarg(c, "collective", &exit);
			MPIB_getopt_coll_optarg(c, operation, timing);
			MPIB_getopt_msg_optarg(c, &stride, &max_size);
			MPIB_getopt_precision_optarg(c, &precision);
		}
	}
	MPIB_getopt_help_bcast(&exit, 0, comm);
	if (exit)
		goto exit;
	MPIB_coll_container* container;
	MPIB_measure_coll measure;
	MPIB_getopt_coll_bcast(operation, &container, timing, &measure, 0, comm);
	if (!container)
	{
		if (rank == 0)
			fprintf(stderr, "Undefined collective operation: %s\n", operation);
		goto exit;
	}
	if (!measure)
	{
		if (rank == 0)
			fprintf(stderr, "Undefined measurement method: %s\n", operation);
		goto exit;
	}
	MPIB_getopt_msg_bcast(&stride, &max_size, 0, comm);
	MPIB_getopt_precision_bcast(&precision, 0, comm);

	MPIB_print_processors(comm);

	int size;
	MPI_Comm_size(comm, &size);

	if (strstr(operation, "MPIB_Scatter_") || strstr(operation, "MPIB_Gather_"))
	{
		if (rank == 0)
		{
			printf("#Test of %s\n", operation);
			printf("#root\tresult\n");
		}
		int root;
		for (root = 0; root < size; root++)
		{
			int res;
			if (strstr(operation, "MPIB_Scatter_"))
			{
				MPIB_Scatter* scatter = (MPIB_Scatter*)((void*)container + sizeof(MPIB_coll_container) + sizeof(char*));
				MPIB_test_scatter(*scatter, comm, root, &res);
			}
			if (strstr(operation, "MPIB_Gather_"))
			{
				MPIB_Gather* gather = (MPIB_Gather*)((void*)container + sizeof(MPIB_coll_container) + sizeof(char*));
				MPIB_test_gather(*gather, comm, root, &res);
			}
			if (rank == 0)
				printf("#%d\t%d\n", root, res);
		}
		if (rank == 0)
			printf("#\n");
	}

	if (rank == 0)
		MPIB_print_coll_th(operation, timing, size, 0, precision);
	double time = 0;
	int M;
	for (M = 0; M < max_size; M += stride)
	{
		MPIB_result result;
		int start = MPI_Wtime();
		measure(container, comm, 0, M, precision, &result);
		time += MPI_Wtime() - start;
		if (rank == 0)
			MPIB_print_coll_tr(M, result);
	}
	if (rank == 0)
		printf("#total benchmarking time: %le\n", time);

	container->free(container);

#ifndef MPIB_DEBUG
	MPI_Comm_free(&comm);
#endif
exit:
	MPI_Finalize();
	return 0;
}
