/*!
 * \file
 * \brief p2p benchmark
 * \details Benchmarks p2p communications between all pairs 
 * 
 * \b p2p.plot draws the graph of the execution time (sec) against message size (kb).
 * - input: p2p.out
 * - output: 0-1.eps (0-1 with error bars), 0-1-2.eps (comparison of 0-1, 0-2, 1-2)
 */

#include "mpib.h"
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

int main(int argc, char** argv)
{
	MPI_Init(&argc, &argv);
	MPI_Comm comm = MPI_COMM_WORLD;
#ifndef MPIB_DEBUG
	MPIB_Comm(MPI_COMM_WORLD, &comm);
	if (comm == MPI_COMM_NULL)
		goto exit;
#endif

	int exit;
	MPIB_getopt_help_default(&exit);
	int stride;
	int max_size;
	MPIB_getopt_msg_default(&stride, &max_size);
	int parallel;
	MPIB_getopt_p2p_default(&parallel);
	MPIB_precision precision;
	MPIB_getopt_precision_default(&precision);

	int rank;
	MPI_Comm_rank(comm, &rank);
	if (rank == 0)
	{
		int c;
		char options[256] = "i:";
		strcat(options, MPIB_getopt_help_options());
		strcat(options, MPIB_getopt_coll_options());
		strcat(options, MPIB_getopt_msg_options());
		strcat(options, MPIB_getopt_p2p_options());
		strcat(options, MPIB_getopt_precision_options());
		while ((c = getopt(argc, argv, options)) != -1)
		{
			MPIB_getopt_help_optarg(c, "p2p", &exit);
			MPIB_getopt_msg_optarg(c, &stride, &max_size);
			MPIB_getopt_p2p_optarg(c, &parallel);
			MPIB_getopt_precision_optarg(c, &precision);
		}
	}
	MPIB_getopt_help_bcast(&exit, 0, comm);
	if (exit)
		goto exit;
	int size;
	MPI_Comm_size(comm, &size);
	if (size < 2)
	{
		if (rank == 0)
			fprintf(stderr, "Cannot perform p2p benchmarks for %d processors (must be >= 2)\n", size);
		goto exit;
	}
	MPIB_getopt_msg_bcast(&stride, &max_size, 0, comm);
	MPIB_getopt_p2p_bcast(&parallel, 0, comm);
	MPIB_getopt_precision_bcast(&precision, 0, comm);

	MPIB_p2p_container* container = MPIB_Send_Recv_container_alloc();
	MPIB_result* results = (MPIB_result*)malloc(sizeof(MPIB_result) * MPIB_C2(size));

	MPIB_print_processors(comm);

	if (rank == 0)
		MPIB_print_p2p_th(parallel, precision, size);
	double time = 0;
	int M;
	for (M = 0; M < max_size; M += stride)
	{
		int start = MPI_Wtime();
		MPIB_measure_p2p(container, comm, M, parallel, precision, results);
		time += MPI_Wtime() - start;
		if (rank == 0)
			MPIB_print_p2p_tr(M, size, results);
	}
	if (rank == 0)
		printf("#total benchmarking time: %le\n\n\n", time);

	M = max_size / 2; 
	MPIB_measure_p2p(container, comm, M, parallel, precision, results);
	if (rank == 0)
		MPIB_print_p2p_table(M, parallel, precision, size, results);

	free(results);
	container->free(container);

#ifndef MPIB_DEBUG
	MPI_Comm_free(&comm);
#endif
exit:
	MPI_Finalize();
	return 0;
}
