/*!
 * \file
 * \brief Collective benchmark
 * \details Checks and benchmarks collective operations, including model-based ones.
 * 
 * Plot can be made by \b MPIBlib/collective/collective.plot
 */

#include "cpm.h"
#include <string.h>
#include <getopt.h>
#include <stdio.h>

int main(int argc, char** argv)
{
	MPI_Init(&argc, &argv);
	MPI_Comm comm = MPI_COMM_WORLD;
#ifndef MPIB_DEBUG
	MPIB_Comm(MPI_COMM_WORLD, &comm);
	if (comm == MPI_COMM_NULL)
		goto exit;
#endif

	int exit;
	MPIB_getopt_help_default(&exit);
	char operation[256];
	char timing[256];
	MPIB_getopt_coll_default(operation, timing);
	int stride;
	int max_size;
	MPIB_getopt_msg_default(&stride, &max_size);
	int parallel;
	MPIB_getopt_p2p_default(&parallel);
	MPIB_precision precision;
	MPIB_getopt_precision_default(&precision);
	char in[256];
	int build = 1;

	int rank;
	MPI_Comm_rank(comm, &rank);
	if (rank == 0)
	{
		int c;
		char options[256] = "i:";
		strcat(options, MPIB_getopt_help_options());
		strcat(options, MPIB_getopt_coll_options());
		strcat(options, MPIB_getopt_msg_options());
		strcat(options, MPIB_getopt_p2p_options());
		strcat(options, MPIB_getopt_precision_options());
		while ((c = getopt(argc, argv, options)) != -1)
		{
			MPIB_getopt_help_optarg(c, "collective", &exit);
			CPM_getopt_coll_optarg(c, operation, timing);
			MPIB_getopt_msg_optarg(c, &stride, &max_size);
			MPIB_getopt_p2p_optarg(c, &parallel);
			MPIB_getopt_precision_optarg(c, &precision);
			switch (c)
			{
				case 'h':
					fprintf
					(
						stderr,
"-i S			input model file, stdin - standard input (default: none - model will be built)\n"
					);
					break;
				case 'i':
					strcpy(in, optarg);
					build = 0;
					break;
			}
		}
	}
	MPIB_getopt_help_bcast(&exit, 0, comm);
	if (exit)
		goto exit;
	MPIB_coll_container* container;
	MPIB_measure_coll measure;
	CPM_getopt_coll_bcast(operation, &container, timing, &measure, 0, comm);
	if (!container)
	{
		if (rank == 0)
			fprintf(stderr, "Undefined collective operation: %s\n", operation);
		goto exit;
	}
	if (!measure)
	{
		if (rank == 0)
			fprintf(stderr, "Undefined measurement method: %s\n", operation);
		goto exit;
	}
	MPIB_getopt_msg_bcast(&stride, &max_size, 0, comm);
	MPIB_getopt_p2p_bcast(&parallel, 0, comm);
	MPIB_getopt_precision_bcast(&precision, 0, comm);
	MPI_Bcast(&build, 1, MPI_INT, 0, comm);

	MPIB_print_processors(comm);

	int size;
	MPI_Comm_size(comm, &size);

	if (!strcmp(operation, "CPM_Scatter_linear_opt") || !strcmp(operation, "CPM_Gather_linear_opt"))
	{
		LMO_model* model = NULL;
		if (build)
			LMO_build(comm, stride, max_size, parallel, precision, &model);
		else
		{
			if (rank == 0)
			{
				model = LMO_load(in);
				exit = model == NULL;
			}
		}
		MPI_Bcast(&exit, 1, MPI_INT, 0, comm);
		if (exit)
			goto exit;
		LMO_initialize(comm, model);
		if (rank == 0)
			LMO_free(model);
	}
	if (!strcmp(operation, "CPM_Scatter_binomial_opt") || !strcmp(operation, "CPM_Scatter_binomial_cor") ||
		!strcmp(operation, "CPM_Gather_binomial_opt") || !strcmp(operation, "CPM_Gather_binomial_cor"))
	{
		Hockney_model* model = NULL;
		if (build)
			Hockney_build(comm, max_size, parallel, precision, &model); 
		else
		{
			if (rank == 0)
			{
				model = Hockney_load(in);
				exit = model == NULL;
			}
		}
		MPI_Bcast(&exit, 1, MPI_INT, 0, comm);
		if (exit)
			goto exit;
		Hockney_initialize(comm, model);
		if (rank == 0)
			Hockney_free(model);
	}

	if (strstr(operation, "MPIB_Scatter_") || strstr(operation, "MPIB_Gather_") ||
		strstr(operation, "CPM_Scatter_") || strstr(operation, "CPM_Gather_"))
	{
		if (rank == 0)
		{
			printf("#Test of %s\n", operation);
			printf("#root\tresult\n");
		}
		int root;
		for (root = 0; root < size; root++)
		{
			int res;
			if (strstr(operation, "MPIB_Scatter_") ||
				strstr(operation, "CPM_Scatter_"))
			{
				MPIB_Scatter* scatter = (MPIB_Scatter*)((void*)container + sizeof(MPIB_coll_container) + sizeof(char*));
				MPIB_test_scatter(*scatter, comm, root, &res);
			}
			if (strstr(operation, "MPIB_Gather_") ||
				strstr(operation, "CPM_Gather_"))
			{
				MPIB_Gather* gather = (MPIB_Gather*)((void*)container + sizeof(MPIB_coll_container) + sizeof(char*));
				MPIB_test_gather(*gather, comm, root, &res);
			}
			if (rank == 0)
				printf("#%d\t%d\n", root, res);
		}
		if (rank == 0)
			printf("#\n");
	}

	if (rank == 0)
		MPIB_print_coll_th(operation, timing, size, 0, precision);
	double time = 0;
	int M;
	for (M = 0; M < max_size; M += stride)
	{
		MPIB_result result;
		int start = MPI_Wtime();
		measure(container, comm, 0, M, precision, &result);
		time += MPI_Wtime() - start;
		if (rank == 0)
			MPIB_print_coll_tr(M, result);
	}
	if (rank == 0)
		printf("#total benchmarking time: %le\n", time);

	if (!strcmp(operation, "CPM_Scatter_linear_opt") || !strcmp(operation, "CPM_Gather_linear_opt"))
		LMO_finalize(comm);

	if (!strcmp(operation, "CPM_Scatter_binomial_opt") || !strcmp(operation, "CPM_Scatter_binomial_cor") ||
		!strcmp(operation, "CPM_Gather_binomial_opt") || !strcmp(operation, "CPM_Gather_binomial_cor"))
		Hockney_finalize(comm);

	container->free(container);

#ifndef MPIB_DEBUG
	MPI_Comm_free(&comm);
#endif
exit:
	MPI_Finalize();
	return 0;
}
