/*!
 * \file
 * \brief Builds the LMO, Hockney and PLogP models and estimates the execution time of ppoint-to-point and
 * collective communication operations.
 * 
 * \b model.plot draws the graph observation vs predictions.
 * - input: lmo.out, hockney.out, plogp.out (model output for different models), p2p.out (p2p output),
 * scatter.out and gather.out (collective outputs for MPIB_Scatter_linear and MPIB_Gather_linear)
 * - output: p2p.eps (0-1), scatter.eps, gather.eps
 */

#include "cpm.h"
#include <string.h>
#include <getopt.h>
#include <stdio.h>

int main(int argc, char** argv)
{
	MPI_Init(&argc, &argv);
	MPI_Comm comm = MPI_COMM_WORLD;
#ifndef MPIB_DEBUG
	MPIB_Comm(MPI_COMM_WORLD, &comm);
	if (comm == MPI_COMM_NULL)
		goto exit;
#endif

	int exit;
	MPIB_getopt_help_default(&exit);
	int stride;
	int max_size;
	MPIB_getopt_msg_default(&stride, &max_size);
	int parallel;
	MPIB_getopt_p2p_default(&parallel);
	MPIB_precision precision;
	MPIB_getopt_precision_default(&precision);
	char model_name[256] = "";
	void* model = NULL;
	char in[256];
	int build = 1;
	char out[256] = "stdout";

	int rank;
	MPI_Comm_rank(comm, &rank);
	if (rank == 0)
	{
		int c;
		char options[256] = "M:i:o:";
		strcat(options, MPIB_getopt_help_options());
		strcat(options, MPIB_getopt_msg_options());
		strcat(options, MPIB_getopt_p2p_options());
		strcat(options, MPIB_getopt_precision_options());
		while ((c = getopt(argc, argv, options)) != -1)
		{
			MPIB_getopt_help_optarg(c, "model", &exit);
			MPIB_getopt_msg_optarg(c, &stride, &max_size);
			MPIB_getopt_p2p_optarg(c, &parallel);
			MPIB_getopt_precision_optarg(c, &precision);
			switch (c)
			{
				case 'h':
					fprintf
					(
						stderr,
"-M S			model name (required): LMO, Hockney, PLogP\n\
-i S			input model file, stdin - standard input (default: none - model will be built)\n\
-o S			output model file, stdout - standard output (default: stdout)\n"
					);
					break;
				case 'M':
					strcpy(model_name, optarg);
					break;
				case 'i':
					strcpy(in, optarg);
					build = 0;
					break;
				case 'o':
					strcpy(out, optarg);
					break;
			}
		}
	}
	MPIB_getopt_help_bcast(&exit, 0, comm);
	if (exit)
		goto exit;
	MPIB_getopt_msg_bcast(&stride, &max_size, 0, comm);
	MPIB_getopt_p2p_bcast(&parallel, 0, comm);
	MPIB_getopt_precision_bcast(&precision, 0, comm);

	int len = strlen(model_name) + 1;
	MPI_Bcast(&len, 1, MPI_INT, 0, comm);
	MPI_Bcast(&model_name, len, MPI_CHAR, 0, comm);
	MPI_Bcast(&build, 1, MPI_INT, 0, comm);
	if (build)
	{
		double time;
		if (!strcmp(model_name, "LMO"))
		{
			if (rank == 0)
			{
				exit = CPM_initR();
				if (exit)
					fprintf(stderr, "R cannot be initialized\n");
			}
			MPI_Bcast(&exit, 1, MPI_INT, 0, comm);
			if (exit)
				goto exit;
	
			time = MPI_Wtime();
			LMO_build(comm, stride, max_size, parallel, precision, (LMO_model**)&model);
			time = MPI_Wtime() - time;
			if (rank == 0)
				CPM_endR();
		}
		if (!strcmp(model_name, "Hockney"))
		{
			time = MPI_Wtime();
			Hockney_build(comm, max_size, parallel, precision, (Hockney_model**)&model); 
			time = MPI_Wtime() - time;
		}
		if (!strcmp(model_name, "PLogP"))
		{
			time = MPI_Wtime();
			PLogP_build(comm, stride, max_size, parallel, precision, (PLogP_model**)&model); 
			time = MPI_Wtime() - time;
		}
		if (rank == 0)
		{
			exit = model == NULL;
			if (exit)
				fprintf(stderr, "Undefined model: %s\n", model_name);
		}
		MPI_Bcast(&exit, 1, MPI_INT, 0, comm);
		if (exit)
			goto exit;
		MPIB_print_processors(comm);
		if (rank == 0)
		{
			if (!strcmp(model_name, "LMO"))
				LMO_save(out, model);
			if (!strcmp(model_name, "Hockney"))
				Hockney_save(out, model);
			if (!strcmp(model_name, "PLogP"))
				PLogP_save(out, model);
			printf("#%s building time: %le\n#\n", model_name, time);
		}
	}
	else
	{
		if (rank == 0)
		{
			if (!strcmp(model_name, "LMO"))
				model = LMO_load(in);
			if (!strcmp(model_name, "Hockney"))
				model = Hockney_load(in);
			if (!strcmp(model_name, "PLogP"))
				model = PLogP_load(in);
			exit = model == NULL;
		}
		MPIB_getopt_help_bcast(&exit, 0, comm);
		if (exit)
			goto exit;
	}

	if (rank == 0)
	{
		if (!strcmp(model_name, "LMO"))
		{
			LMO_model* lmo = (LMO_model*)model;
			int size;
			MPI_Comm_size(comm, &size);
			printf("#LMO prediction\n#msg\t");
			int i;
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("%d-%d\t\t", i, j);
			}
			printf("\n");
			printf("0");
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", LMO_estimate_p2p(lmo, i, j, 0) * 2);
			}
			printf("\n");
			printf("%d", max_size);
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", LMO_estimate_p2p(lmo, i, j, max_size) * 2);
			}
			printf("\n\n\n");
			int M;
			printf("#scatter prediction\n");
			printf("#msg\tprediction\n");
			M = 0;
			printf("%d\t%le\n", M, LMO_estimate_one2many(lmo, 0, M));
			M = lmo->S - 1;
			printf("%d\t%le\n", M, LMO_estimate_one2many(lmo, 0, M));
			M = lmo->S + 1;
			printf("%d\t%le\n", M, LMO_estimate_one2many(lmo, 0, M));
			M = max_size;
			printf("%d\t%le\n", M, LMO_estimate_one2many(lmo, 0, M));
			printf("\n\n");
			printf("#gather prediction\n");
			printf("#msg\tprediction\n");
			M = 0;
			printf("%d\t%le\n", M, LMO_estimate_many2one(lmo, 0, M));
			M = lmo->M2 - 1;
			printf("%d\t%le\n", M, LMO_estimate_many2one(lmo, 0, M));
			M = lmo->M2 + 1;
			printf("%d\t%le\n", M, LMO_estimate_many2one(lmo, 0, M));
			M = max_size;
			printf("%d\t%le\n", M, LMO_estimate_many2one(lmo, 0, M));
			printf("\n");
		}
		if (!strcmp(model_name, "Hockney"))
		{
			Hockney_model* hockney = (Hockney_model*)model;
			printf("#Hockney prediction\n#msg\t");
			int size;
			MPI_Comm_size(comm, &size);
			int i;
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("%d-%d\t\t", i, j);
			}
			printf("\n");
			printf("0");
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", Hockney_estimate(hockney, i, j, 0) * 2);
			}
			printf("\n");
			printf("%d", max_size);
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", Hockney_estimate(hockney, i, j, max_size) * 2);
			}
			printf("\n\n\n");
			printf("#Hockney_estimate_sg_linear_sequential\n");
			printf("#msg\tprediction\n");
			printf("%d\t%le\n", 0, Hockney_estimate_sg_linear_sequential(hockney, 0, 0));
			printf("%d\t%le\n", max_size, Hockney_estimate_sg_linear_sequential(hockney, 0, max_size));
			printf("\n\n");
			printf("#Hockney_estimate_sg_linear_parallel\n");
			printf("#msg\tprediction\n");
			int M;
			for (M = 0; M < max_size; M += stride)
				printf("%d\t%le\n", M, Hockney_estimate_sg_linear_parallel(hockney, 0, M));
			printf("\n\n");
			printf("#Hockney_estimate_sg_binomial\n");
			printf("#msg\tprediction\n");
			for (M = 0; M < max_size; M += stride)
				printf("%d\t%le\n", M, Hockney_estimate_sg_binomial(hockney, 0, M));
			printf("\n\n");
			printf("#Hockney_hestimate_sg_linear_sequential\n");
			printf("#msg\tprediction\n");
			printf("%d\t%le\n", 0, Hockney_hestimate_sg_linear_sequential(hockney, 0, 0));
			printf("%d\t%le\n", max_size, Hockney_hestimate_sg_linear_sequential(hockney, 0, max_size));
			printf("\n\n");
			printf("#Hockney_hestimate_sg_linear_parallel\n");
			printf("#msg\tprediction\n");
			printf("%d\t%le\n", 0, Hockney_hestimate_sg_linear_parallel(hockney, 0, 0));
			printf("%d\t%le\n", max_size, Hockney_hestimate_sg_linear_parallel(hockney, 0, max_size));
			printf("\n\n");
			printf("#Hockney_hestimate_sg_binomial\n");
			printf("#msg\tprediction\n");
			printf("%d\t%le\n", 0, Hockney_hestimate_sg_binomial(hockney, 0, 0));
			printf("%d\t%le\n", max_size, Hockney_hestimate_sg_binomial(hockney, 0, max_size));
		}
		if (!strcmp(model_name, "PLogP"))
		{
			PLogP_model* plogp = (PLogP_model*)model; 
			printf("#PLogP prediction\n#msg\t");
			int size;
			MPI_Comm_size(comm, &size);
			int i;
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("%d-%d\t\t", i, j);
			}
			printf("\n");
			int M;
			for (M = 0; M < max_size; M += stride)
			{
				printf("%d", M);
				for(i = 0; i < size - 1; i++)
				{
					int j;
					for(j = i + 1; j < size; j++)
						printf("\t%le", PLogP_estimate(plogp, i, j, M) * 2);
				}
				printf("\n");
			}
			printf("\n\n");
			printf("#LogGP prediction\n#msg\t");
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("%d-%d\t\t", i, j);
			}
			printf("\n");
			printf("0");
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", PLogP_estimate_LogGP(plogp, i, j, 0) * 2);
			}
			printf("\n");
			printf("%d", max_size);
			for(i = 0; i < size - 1; i++)
			{
				int j;
				for(j = i + 1; j < size; j++)
					printf("\t%le", PLogP_estimate_LogGP(plogp, i, j, max_size) * 2);
			}
			printf("\n");
		}
	}

	if (!strcmp(model_name, "LMO"))
		LMO_free((LMO_model*)model);
	if (!strcmp(model_name, "Hockney"))
		Hockney_free((Hockney_model*)model);
	if (!strcmp(model_name, "PLogP"))
		PLogP_free((PLogP_model*)model);

#ifndef MPIB_DEBUG
	MPI_Comm_free(&comm);
#endif
exit:
	MPI_Finalize();
	return 0;
}
