#ifndef CPM_H_
#define CPM_H_

/*!
 * \mainpage Introduction
Traditionally, communication performance models for high performance computing are analytical and built for homogeneous clusters. The basis of these models is a point-to-point communication model characterized by a set of integral parameters, having the same value for each pair of processors. The execution time of other operations (which are, in fact, collective), is expressed as a combination of the point-to-point parameters, and is analytically predicted for different message sizes and numbers of processors involved. The core of this approach is the choice of such a point-to-point model that is the most appropriate to the targeted platform, allowing for easy and natural expression of different algorithms of collective operations. For homogeneous clusters, the point-to-point parameters are found statistically from communication experiments between any two processors. Typical experiments include sending and receiving messages of different sizes, with the communication execution time being measured on one side. 

A homogeneous communication model can be applied to a cluster of heterogeneous processors by averaging values obtained for every pair of processors. In this case, the heterogeneous cluster will be treated as homogeneous in terms of the performance of communication operations. If some processors or links in the heterogeneous cluster significantly differ in performance, predictions based on the homogeneous communication model may become inaccurate. More accurate performance models would not average the point-to-point communication parameters. The use of such heterogeneous communication models in model-based optimization of MPI collective operations on heterogeneous clusters do improve their performance.

The traditional models use a small number of parameters to describe communication between any two processors. The number of these parameters and their use in the model are always defined in a way that allows for their accurate estimation with a set of point-to-point communication experiments between these two processors. The price to pay is that such a traditional point-to-point communication model is not intuitive. The meaning of its parameters is not clear. Different sources of the contribution into the execution time are artificially and non-intuitively mixed and spread over a smaller number of parameters. This makes the models difficult to use for accurate modelling of collective communications.

The alternative approach is to use original point-to-point heterogeneous models that allow for easy and intuitive expression of the execution time of collective communication operations such as this model designed for switched heterogeneous clusters. While more accurate, this heterogeneous model has a significantly larger number of parameters. This will result in a higher cost of their estimation. In particular, when applied to the heterogeneous communication model, the statistical methods of finding the point-to-point parameters, traditionally used in the case of homogeneous communication models, will require a significantly larger number of measurements. For our target architecture, which is a heterogeneous cluster based on a switched network, we can address this problem by performing most of the communication experiments in parallel, using the fact that the network switches provide no-contention point-to-point communications, appropriately forwarding packets between sources and destinations. 
 * 
 * \section authors Authors
 * \include AUTHORS
 * 
 * \section Changes
 * \include NEWS
 * 
 * \page installation Installation
 * \include README
 * 
 * \page usage Usage
 * The package consists of a library, executables and gnuplot scripts.
 * 
 * \dotfile src/cpm.dot
 * 
 * Library structure:
 * 
 * \dotfile src/cpm_lib.dot
 *
 * Benchmarking executables and gnuplot scripts are described in:
 * - \ref model/main.c
 * - \ref collective/main.c
 * 
 * Typical parameters of executables are listed in \ref getopt.
 * 
 * Using the gnuplot
 * \code
 * $ gnuplot script_name.plot
 * \endcode
 * The gnuplot data files should have names script_name.out
 */

#include "mpib.h"

#include "cpm_config.h"

#include "cpm_utilities.h"
#include "cpm_measurement.h"
#include "cpm_lmo.h"
#include "cpm_hockney.h"
#ifdef HAVE_LIBLOGP
#include "cpm_plogp.h"
#endif /*HAVE_LIBLOGP*/
#include "cpm_collective.h"
#include "cpm_getopt.h"

#endif /*CPM_H_*/
