#ifndef CPM_HOCKNEY_H_
#define CPM_HOCKNEY_H_

/*!
 * \defgroup hockney Hockney
 * \brief Computes the parameters of the Hockney model.
 * \{
 */

/*!
 * \brief Hockney model
 * \details \f$ T = \alpha + \beta M \f$
 */
typedef struct Hockney_model
{
	/*! number of nodes */
	int n;
	/*! array of \f$ C_n^2 \f$ of \f$ \alpha \f$ */
	double* a;
	/*! array of \f$ C_n^2 \f$ of \f$ \beta \f$ */
	double* b;
}
Hockney_model;

/*!
 * \brief Builds the Hockney model.  
 * \details
 * - Measures the execution time \f$ T_{ij}(0) \f$
 * of each \f$ i \xlongleftrightarrow[0]{0} j \f$ roundtrip in the communicator, \f$ i < j \f$,
 * to find \f$ \alpha = T_{ij}(0) \f$. To obtain more accurate results performs a series of roundtrips
 * and takes the average \f$ T_{ij}(0) \f$.
 * - Measures the execution time \f$ T_{ij}(M) \f$
 * of each \f$ i \xlongleftrightarrow[M]{M} j \f$ roundtrip in the communicator, \f$ i < j \f$,
 * to find \f$ \beta = \displaystyle\frac{T_{ij}(M) - \alpha}{M} \f$. To obtain more accurate results performs a series of roundtrips
 * and takes the average \f$ T_{ij}(M) \f$.
 * \param comm communicator, number of nodes \f$ \ge 2 \f$
 * \param M message size
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param precision measurement precision
 * \param model Hockney model (significant only at root)
 */
void Hockney_build(MPI_Comm comm, int M, int parallel, MPIB_precision precision, Hockney_model** model);

/*!
 * \brief Allocates memory for Hockney model.
 */
Hockney_model* Hockney_alloc(int n);

/*!
 * \brief Frees Hockney model.
 * \param model Hockney model
 */
void Hockney_free(Hockney_model* model);

/*!
 * \brief Loads Hockney model.
 * \param filename file name ("stdin" = stdin)
 * \return Hockney model
 */
Hockney_model* Hockney_load(const char* filename);

/*!
 * \brief Saves Hockney model.
 * \param filename file name ("stdout" = stdout)
 * \param model Hockney model
 */
void Hockney_save(const char* filename, Hockney_model* model);

/*!
 * \brief Estimates the execution time of a point-to-point communication.
 */
double Hockney_estimate(Hockney_model* model, int i, int j, int M);

/*!
 * \brief Heterogeneous estimation of linear scatter/gather (sequential point-to-point communications)
 * \details \f$ \displaystyle\sum_{i = 0, i \neq r }^{n - 1}(\alpha_{ri} + \beta_{ri} M) \f$
 */
double Hockney_estimate_sg_linear_sequential(Hockney_model* model, int root, int M);

/*!
 * \brief Heterogeneous estimation of linear scatter/gather (parallel point-to-point communications)
 * \details \f$ \displaystyle\max_{i = 0, i \neq r }^{n - 1}(\alpha_{ri} + \beta_{ri} M) \f$
 */
double Hockney_estimate_sg_linear_parallel(Hockney_model* model, int root, int M);

/*!
 * \brief Heterogeneous estimation of binomial scatter/gather (point-to-point communications are
 * sequential within the same root or parallel otherwise)
 * \details \f$ \displaystyle\overbrace{\sum_{k = log_2(n - 1)}^{0}(\alpha_{ri} + \beta_{ri} 2^k M) +
 * \displaystyle\max_{k = log_2(n - 1) - 1}^{0}(S(k))}^{S(log_2(n - 1))} \f$
 */
double Hockney_estimate_sg_binomial(Hockney_model* model, int root, int M);

/*!
 * \brief Homogeneous estimation of linear scatter/gather (sequential point-to-point communications)
 * \details \f$ (n - 1)(\alpha + \beta M) \f$
 */
double Hockney_hestimate_sg_linear_sequential(Hockney_model* model, int root, int M);

/*!
 * \brief Homogeneous estimation of linear scatter/gather (parallel point-to-point communications)
 * \details \f$ \alpha + \beta M \f$
 */
double Hockney_hestimate_sg_linear_parallel(Hockney_model* model, int root, int M);

/*!
 * \brief Homogeneous estimation of binomial scatter/gather (point-to-point communications are
 * sequential within the same root or parallel otherwise)
 * \details \f$ (log_2 n) \alpha + (n - 1) \beta M \f$
 */
double Hockney_hestimate_sg_binomial(Hockney_model* model, int root, int M);

/*!
 * \}
 */

#endif /*CPM_HOCKNEY_H_*/
